//===========================================================================
// Beispiel zur Simulation von Partikeln
//===========================================================================

//===========================================================================
// includes
//===========================================================================

// GL und GLUT Bibliotheken einbinden (je nach Betriebssystem)

#ifdef WIN32
    #include <windows.h>
    #include <GL/gl.h>
    #include <GL/glut.h>
#else
    #include <GLUT/glut.h>
#endif

#include "texture.h"

#include <math.h>
#include <iostream>
using namespace std;

//===========================================================================
// Konstanten
//===========================================================================

const float PI            = asin(1.0f)*2.0f;

//===========================================================================
// Globale variablen
//===========================================================================

int      glutWindow = 0;

CTexture     original;
CTexture     moved;
CTexture     final;
float        rotation = 0.0f;


//===========================================================================
// Texturfilter
//===========================================================================

// Bild um festen Offset bewegen
void moveFilter(const CTexture& t1, CTexture* t2, int moveX, int moveY)
{
    // Gre des gefilterten Bildes angleichen
    if (!t2->resize(t1.width(), t1.height()))
        return;

    // Hhe und Breite merken (Geschwindigkeit)
    const unsigned int width  = t1.width();
    const unsigned int height = t1.height();

    // Verschiebung in positive Werte umrechnen
    // (Vorsicht vor modulo von negativen Zahlen)
    unsigned int mX, mY;

    if (moveX >= 0)
        mX = moveX;
    else
        mX = (width - (-moveX % width)) % width;

    if (moveY >= 0)
        mY = moveY;
    else
        mY = (height - (-moveY % height)) % height;

    RGBAPixel color;
    for (unsigned int j = 0; j < height; j++)
    {
        unsigned int y = (j+mY) % height;

        for (unsigned int i = 0; i < width; i++)
        {
            unsigned int x = (i+mX) % width;

            // Einzelnen Pixel verschieben
            t1.getPixel(i, j, &color);
            t2->setPixel(x, y, color);
        }
    }

} // moveFilter()


// Bilder ineinander berblenden
bool blendFilter(const CTexture& t1, const CTexture& t2, CTexture* t3, float mix)
{
    // Parameter validieren
    if (t3 == NULL)
        return false;
    if (t1.width() != t2.width() || t1.height() != t2.height())
        return false;

    // Gre des gefilterten Bildes angleichen
    if (!t3->resize(t1.width(), t1.height()))
        return false;

    // Hhe und Breite merken (Geschwindigkeit)
    const unsigned int width  = t1.width();
    const unsigned int height = t1.height();

    RGBAPixel color1, color2, color;
    for (unsigned int j = 0; j < height; j++)
    {
        for (unsigned int i = 0; i < width; i++)
        {
            // Einzelnen Pixel ineinander blenden (mit Rundung)
            t1.getPixel(i, j, &color1);
            t2.getPixel(i, j, &color2);
            color.r = (unsigned char) (float(color1.r) * (1.0f-mix) + float(color2.r) * mix + 0.5f);
            color.g = (unsigned char) (float(color1.g) * (1.0f-mix) + float(color2.g) * mix + 0.5f);
            color.b = (unsigned char) (float(color1.b) * (1.0f-mix) + float(color2.b) * mix + 0.5f);
            color.a = (unsigned char) (float(color1.a) * (1.0f-mix) + float(color2.a) * mix + 0.5f);
            t3->setPixel(i, j, color);
        }
    }

    return true;

} // blendFilter()


//===========================================================================
// OpenGL Zeichenfunktion
//===========================================================================

void transform(unsigned int width, unsigned int height)
{
    // OpenGL Transformationen initialisieren...
    // Ausgabefenstergre setzen
    glViewport(0, 0, (GLint)width, (GLint)height);

    // Verhltnis Breite zu Hhe (aspect) berechnen
    GLdouble aspect = (GLdouble)width / (GLdouble)height;
    
    // Perspektivische Projektion bearbeiten
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // Kameraperspektive setzen
    GLdouble fieldOfView = atan(1.0/2.0)*180.0/double(PI)*2.0;
    gluPerspective(fieldOfView,  // ffnungswinkel vertikal (field of view)
                   aspect,       // aspect ratio setzen
                   0.001, 10.0); // nahe und ferne Begrenzungsebene setzen

    // Transformationen fr 3D-Koordinaten bearbeiten
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // Kameraposition setzen
    gluLookAt(0.0, 0.0, 2.0,  // von wo
              0.0, 0.0, 0.0,  // nach wo
              0.0, 1.0, 0.0); // wo ist oben?
}


void update()
{
    // Rotation innerhalb von 30 Sekunden um 2*PI laufen lassen
    // (angenommen sind 60Hz Wiederholrate)
    rotation += 2.0f*PI/60.0f/30.0f;
    rotation = fmod(rotation, 2.0f*PI);
}


void render()
{
    float aspect = float(glutGet(GLUT_WINDOW_WIDTH)) /
                   float(glutGet(GLUT_WINDOW_HEIGHT));

    // Grad der Bewegung bestimmen
    int moveX = int(cos(rotation)*float(original.width()));
    int moveY = int(sin(rotation)*float(original.height()));

    // Zwei Filter anwenden: Bewegung und berblendung
    moveFilter(original, &moved, moveX, moveY);
    blendFilter(moved, final, &final, 250.0f/255.0f);

    // Textur in die Grafikkarte laden und verwenden
    final.upload();
    final.use();

    // zeichnen
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    glBegin(GL_QUADS);
    glTexCoord2f(0, 0); glVertex2f(-aspect, -1);
    glTexCoord2f(1, 0); glVertex2f( aspect, -1);
    glTexCoord2f(1, 1); glVertex2f( aspect,  1);
    glTexCoord2f(0, 1); glVertex2f(-aspect,  1);
    glEnd();

    // Textur deaktivieren
    final.disable();
}


//===========================================================================
// GLUT callback funktionen
//===========================================================================

void display(void)
{
    // Ausgabefenster in schwarz lschen
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Transformationen vor dem Zeichnen setzen
    transform(glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT));

    // Simulation aktualisieren
    update();

    // Zeichnen
    render();

    // Hintergrundpuffer anzeigen
    glutSwapBuffers();
}


void idle(void)
{
    // nochmaliges Zeichnen einleiten
    if (glutWindow != 0) glutPostRedisplay();
}


void keyboard(unsigned char key, int x, int y)
{
    // Texturen lschen
    final.remove();
    moved.remove();
    original.remove();

    // Fenster schlieen
    glutDestroyWindow(glutWindow);
    glutWindow = 0;

    // Programm beenden (etwas unsauber)
    exit(0);
}


void motion(int x, int y)
{
    // Hier Dinge einfgen, die getan werden sollen,
    // wenn die Maus bewegt wird...
}


void mouse(int button, int state, int x, int y)
{
    if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
    {
        // Hier Dinge einfgen, die getan werden sollen,
        // wenn die linke Maustaste gedrckt wird...
    }
}


//===========================================================================
// Hauptroutine
//===========================================================================

int main(int argc, char* argv[])
{
    // DevIL initialisieren
    ilInit();

    // GLUT initialisieren
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);

    // Fenster ffnen
    glutWindow = glutCreateWindow("texture");

    // GLUT Funktionen fr Zeichnen, Leerlauf, Tastatur und Maus setzen
    glutDisplayFunc(display);
    glutIdleFunc(idle);
    glutKeyboardFunc(keyboard);
    glutMotionFunc(motion);
    glutPassiveMotionFunc(motion);
    glutMouseFunc(mouse);

    // Vollbildmodus ohne Cursor
    glutFullScreen();
    glutSetCursor(GLUT_CURSOR_NONE);

    // Tiefentest einschalten
    glEnable(GL_DEPTH_TEST);

    // Texturen vorbereiten
    original.load("colors.png");
    moved.resize(original.width(), original.height());
    final.resize(original.width(), original.height());
    RGBAPixel black = { 0, 0, 0, 0 };
    moved.clear(black);
    final.clear(black);

    // Hauptschleife des Programms aufrufen
    glutMainLoop();

    // Bis hierhin kommt das Programm nicht
    return 0;
}
