/***************************************************************************
 * @file  texture.c
 * @brief Implementierung zum Darstellen von Texturen mit OpenGL
 ***************************************************************************/

/***************************************************************************
 * Includes
 ***************************************************************************/

#include "texture.h"

#include <memory.h> // NULL


/***************************************************************************
 * Funktionen zum Verndern von texturen
 ***************************************************************************/

CTexture::CTexture()
{
    m_width         = 0;
    m_height        = 0;
    m_buffer        = NULL;
    m_sizeChanged   = true;
    m_textureNumber = 0;
    m_imageNumber   = 0;
}


CTexture::~CTexture()
{
    // Nichts tun, da wir nicht wissen, ob
    // unser OpenGL-Fenster noch aktiv ist...

    // Textur aus DevIL lschen falls ntig
    if (m_imageNumber != 0)
    {
        ilDeleteImages(1, &m_imageNumber);
        m_imageNumber = 0;
    }
}


bool CTexture::generate(const unsigned int width,
                        const unsigned int height)
{
    // Textur wurde noch nicht in Grafikkarte bertragen,
    // deshalb wie in OpenGL blich Texturnummer auf 0 setzen
    m_textureNumber = 0;

    // Speicher fr Textur belegen
    return resize(width, height);

} // generate()


bool CTexture::remove()
{
    // Textur aus Grafikkarte lschen falls ntig
    if (m_textureNumber != 0)
    {
        glDeleteTextures(1, &m_textureNumber);
        m_textureNumber = 0;
    }

    // Speicher freigeben
    if (m_buffer != NULL)
    {
        free(m_buffer);
        m_buffer = NULL;
    }

    // Breite und Hhe zurcksetzen
    m_width  = 0;
    m_height = 0;

    // Alles in Ordnung
    return true;

} // remove()


unsigned int CTexture::width() const
{
    return m_width;

} // width()


unsigned int CTexture::height() const
{
    return m_height;

} // height()


bool CTexture::clear(const RGBAPixel& pixel)
{
    unsigned int i;             // Laufvariable
    RGBAPixel*   buffer;        // Schneller Zugriff auf Pixelspeicher

    // Internen Status validieren
    if (m_buffer == NULL) return false;

    // Zeiger auf Pixelspeicher in Variable laden
    // (schnellerer Zugriff)
    buffer = m_buffer;

    // Pixel rckwrts zhlen (schneller)
    for (i = m_width * m_height; i > 0; --i)
    {
        // Aktuellen Pixel auf bergebene Farbe setzen
        *buffer++ = pixel;
    }

    return true;

} // clear()


bool CTexture::setPixel(const unsigned int x,
                        const unsigned int y,
                        const RGBAPixel& pixel)
{
    // Internen Status validieren
    if (m_buffer == NULL) return false;
    if (x >= m_width    ) return false;
    if (y >= m_height   ) return false;

    // Pixel setzen
    m_buffer[y*m_width + x] = pixel;

    // Alles okay
    return true;

} // setPixel()


bool CTexture::getPixel(const unsigned int x,
                        const unsigned int y,
                        RGBAPixel* pixel) const
{
    // Eingabeparameter und internen Status validieren
    if (pixel    == NULL) return false;
    if (m_buffer == NULL) return false;
    if (x >= m_width    ) return false;
    if (y >= m_height   ) return false;

    // Pixel lesen
    *pixel = m_buffer[y*m_width + x];

    // Alles okay
    return true;

} // getPixel()


bool CTexture::upload()
{
    // OpenGL Texturen aktivieren
    glEnable(GL_TEXTURE_2D);

    // Textur in Grafikkarte anlegen, falls noch nicht existent
    if (m_textureNumber == 0)
    {
        // Neue texturnummer erzeugen
        glGenTextures(1, &m_textureNumber);

        // Existiert jetzt eine Textur in der Grafikkarte?
        if (m_textureNumber == 0) return false;

        // OpenGL Texturziel auf unsere Textur setzen
        glBindTexture(GL_TEXTURE_2D, m_textureNumber);

        // Lineare Interpolation der Texturpixel verwenden
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        // Textur wird am Rand nicht wiederholt
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

        // Lineare Objekt-Texturkoordinaten verwenden
        glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);

        // Textur auf Farbmodulation beleuchteter Flchen setzen
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

        // Textur neu hochladen
        glTexImage2D(GL_TEXTURE_2D,     // 2D Textur
                     0,                 // Detailsstufe (fr Mipmaps)
                     4,                 // Farbkomponenten (4 fr RGBA)
                     m_width,           // Breite
                     m_height,          // Hhe
                     0,                 // Rand
                     GL_RGBA,           // Pixel-Format (RGBA)
                     GL_UNSIGNED_BYTE,  // Datentyp der Komponenten (0 bis 255)
                     m_buffer);         // Pixel-Puffer
    }
    else if (m_sizeChanged)
    {
        // existente Textur neuer Gre aktualisieren

        // OpenGL Texturziel auf unsere Textur setzen
        glBindTexture(GL_TEXTURE_2D, m_textureNumber);

        // Textur neu hochladen
        glTexImage2D(GL_TEXTURE_2D,     // 2D Textur
                     0,                 // Detailsstufe (fr Mipmaps)
                     4,                 // Farbkomponenten (4 fr RGBA)
                     m_width,           // Breite
                     m_height,          // Hhe
                     0,                 // Rand
                     GL_RGBA,           // Pixel-Format (RGBA)
                     GL_UNSIGNED_BYTE,  // Datentyp der Komponenten (0 bis 255)
                     m_buffer);         // Pixel-Puffer

        m_sizeChanged = FALSE;
    }
    else
    {
        // existente Textur aktualisieren

        // OpenGL Texturziel auf unsere Textur setzen
        glBindTexture(GL_TEXTURE_2D, m_textureNumber);

        // Texturspeicher auf Grafikkarte aktualsieren
        glTexSubImage2D(GL_TEXTURE_2D,     // 2D Textur
                        0,                 // Detailsstufe (fr Mipmaps)
                        0,                 // X offset
                        0,                 // Y offset
                        m_width,           // Breite
                        m_height,          // Hhe 
                        GL_RGBA,           // Pixel-Format (RGBA)
                        GL_UNSIGNED_BYTE,  // Datentyp der Komponenten (0 bis 255)
                        m_buffer);         // Pixel-Puffer
    }

    // Verwenden von Texturen wieder ausschalten
    glDisable(GL_TEXTURE_2D);

    // Alles in Ordnung
    return true;

} // upload()


bool CTexture::use()
{
    // Eingabeparameter validieren
    if (m_textureNumber == 0) return false;

    // OpenGL Texturen aktivieren
    glEnable(GL_TEXTURE_2D);

    // Alpha-Kanal aktivieren
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // Texturoperationen auf unsere Textur setzen
    glBindTexture(GL_TEXTURE_2D, m_textureNumber);

    return true;

} // use()


bool CTexture::disable()
{
    glDisable(GL_TEXTURE_2D);

    return true;

} // disable()


bool CTexture::resize(const unsigned int width,
                      const unsigned int height)
{
    if (width*height == 0)
    {
        // Speicher freigeben
        if (m_buffer != NULL)
        {
            delete[] m_buffer;
            m_buffer = NULL;
        }

        // Breite und Hhe zurcksetzen
        m_width  = 0;
        m_height = 0;
    }
    else if (m_width != width || m_height != height)
    {
        // Neuen Speicher belegen
        RGBAPixel* newBuffer = new RGBAPixel[width * height];

        // War Speicherbelegung erfolgreich?
        if (newBuffer == NULL) return false;

        // Alten Puffer beseitigen
        if (m_buffer != NULL) delete[] m_buffer;

        // Neue Werte bernehmen
        m_buffer      = newBuffer;
        m_width       = width;
        m_height      = height;
        m_sizeChanged = true;
    }

    // Alles okay
    return true;
}


bool CTexture::load(const char* filename)
{
    // Eingabeparameter validieren
    if (filename == NULL) return false;

    // Bild in DevIL anlegen
    if (m_imageNumber == 0)
      ilGenImages(1, &m_imageNumber);

    // DevIL mitteilen, dass mit unserem Bild gearbeitet werden soll
    ilBindImage(m_imageNumber);

    // Versuchen, das Bild zu laden
    if (!ilLoadImage((const ILstring) filename)) return false;

    // Textur in neuer Breite und Hhe anlegen
    if (!resize(ilGetInteger(IL_IMAGE_WIDTH),
                ilGetInteger(IL_IMAGE_HEIGHT))) return false;

    // Puffer in unser eigenes Bild kopieren
    if (!ilCopyPixels(0,                // X Versatz
                      0,                // Y Versatz
                      0,                // Z Versatz
                      m_width,          // Breite
                      m_height,         // Hhe
                      1,                // Tiefe
                      IL_RGBA,          // Zielformat
                      IL_UNSIGNED_BYTE, // Komponenten
                      m_buffer)         // Texturpuffer
       ) return false;

    // Pixel vertikal spiegeln
    for (unsigned int j = 0; j < m_height/2; j++)
    {
        for (unsigned int i = 0; i < m_width; i++)
        {
            RGBAPixel color1, color2;

            getPixel(i, j,            &color1);
            getPixel(i, m_height-1-j, &color2);

            setPixel(i, j,            color2);
            setPixel(i, m_height-1-j, color1);
        }
    }

    // Alles okay
    return true;
}
