/***************************************************************************
 * @file  texture.h
 * @brief Interface zum Darstellen von Texturen mit OpenGL
 ***************************************************************************/

// mehrfaches Einbinden dieser Header-Datei verhindern
#ifndef _TEXTURE_H_
#define _TEXTURE_H_


/***************************************************************************
 * Includes
 ***************************************************************************/

// GL und GLUT Bibliotheken einbinden (je nach Betriebssystem)
#ifdef WIN32
    #include <windows.h>
    #include <GL/gl.h>
    #include <GL/glut.h>
#else
    #include <GLUT/glut.h>
#endif

// DevIL Bibliothek einbinden
#include <IL/il.h>


/***************************************************************************
 * Typendeklarationen
 ***************************************************************************/

typedef struct
{
    unsigned char r;    ///< Rote Farbkomponente
    unsigned char g;    ///< Grne Farbkomponente
    unsigned char b;    ///< Blaue Farbkomponente
    unsigned char a;    ///< Transparenzkomponente

} RGBAPixel;


/***************************************************************************
 * Funktionen zum Verndern von texturen
 ***************************************************************************/

class CTexture
{
  public:

    CTexture();
    virtual ~CTexture();

    /**
     * @brief Erzeugen einer Textur
     */
    bool generate(const unsigned int width,
                  const unsigned int height);

    /**
     * @brief Lschen einer vorher generierten Textur
     */
    bool remove();

    /**
     * @brief Breite der Textur zurckgeben
     */
    unsigned int width() const;

    /**
     * @brief Hhe der Textur zurckgeben
     */
    unsigned int height() const;

    /**
     * @brief Textur mit einer Farbe lschen
     */
    bool clear(const RGBAPixel& pixel);

    /**
     * @brief Pixel an Position (x,y) in Textur setzen
     */
    bool setPixel(const unsigned int x,
                  const unsigned int y,
                  const RGBAPixel& pixel);

    /**
     * @brief Pixel an Position (x,y) aus Textur lesen
     */
    bool getPixel(const unsigned int x,
                  const unsigned int y,
                  RGBAPixel* pixel) const;

    /**
     * @brief Hochladen einer Textur in die Grafikkarte
     */
    bool upload();

    /**
     * @brief Zeichnen mit der textur initialisieren
     */
    bool use();

    /**
     * @brief Zeichnen mit der textur beenden
     */
    bool disable();

    /**
     * @brief Verndern der Gre einer Textur
     *
     * Vorsicht: Das Bild wird nicht skaliert. Nur der interne Puffer
     * wird in seiner Gre angepasst. Zum Initialisieren muss
     * clear() aufgerufen werden.
     *
     * @see   clear()
     */
    bool resize(const unsigned int width,
                const unsigned int height);

    /**
     * @brief Textur aus Datei laden
     */
    bool load(const char* filename);

  protected:

    unsigned int m_width, m_height; ///< Breite und Hhe der Textur
    RGBAPixel*   m_buffer;          ///< Speicherpuffer fr Pixel der Textur
    bool         m_sizeChanged;     ///< Gre der Textur hat sich verndert
    GLuint       m_textureNumber;   ///< OpenGL Texturnummer
    ILuint       m_imageNumber;     ///< DevIL Bildnummer
};

#endif // _TEXTURE_H_ (mehrfaches Einbinden verhindern)
